#!/bin/bash

######################################################################
# Proprietary Information D-Wave Systems Inc.                        #
# Copyright (c) 2017 by D-Wave Systems Inc. All rights reserved.     #
# D-Wave Systems Inc., 3033 Beta Ave., Burnaby, BC, V5G 4M9, Canada. #
######################################################################

# The script handles two special types of constraints: one is
# assertions that are weighted linear sums of variables and the other
# is assertions that are QUBOs.  The script reads the A**** assertion
# file so that it has access to the variable names.  It then reads the
# A****.coefficient file which has the model assertion as well as an
# indicator.  If the indicator LINEAR is present, then we can create a
# QUBO by squaring the linear constraint.  If the indicator QUBO is
# present, then we can simply pass on the assertion since it is a
# valid QUBO.  The output of this script is an A****.posiform file.
# If this script exits with status 0, then it has found one of these
# two special cases and has generated an A****.posiform file.
# Otherwise, the script exits with status 1 and writes no additional
# output.

awk '

BEGIN { debug = 0; varstr = ""; exit_status = 1; }

$1 ~ /^active_vars/ {
    split($0, token, "=");
    sub(/.*\[/,"",token[1]);
    sub(/\]/,"",token[1]);
    active_vars[token[1]] = token[2];
    if (0 + token[1] > 0 + max_var) max_var = token[1];
    varstr = varstr "2";
}

$1 ~ /^constant/ { constant = $2 }

$1 ~ /^linear/ { sub(/.*\[/,"",$1); sub(/\]:/,"",$1); linear[$1] = $2; }

$1 ~ /^quadratic/ { sub(/.*\[/,"",$1); sub(/\]:/,"",$1); quadratic[$1] = $2; }

$1 ~ /^pair/ { sub(/.*\[/,"",$1); sub(/\]:/,"",$1); pair[$1] = $2; }

$1 ~ /^LINEAR$/ {

    term_wt[""] = constant * constant;

    for (i=0; i<=max_var; ++i)
        term_wt[i] = linear[i]*linear[i] + 2*linear[i]*constant;

    for (i=0; i<=max_var; ++i)
        for (j=i+1; j<=max_var; ++j)
            term_wt[i,j] = 2*linear[i]*linear[j];

    display();

    exit_status = 0;
}

$1 ~ /^QUBO$/ {

    term_wt[""] = constant;

    for (i=0; i<=max_var; ++i)
        term_wt[i] = linear[i] + quadratic[i];

    for (i=0; i<=max_var; ++i)
        for (j=i+1; j<=max_var; ++j)
            term_wt[i,j] = pair[i "," j];

    display();

    exit_status = 0;
}

function display()
{
    for (i=0; i<=max_var; ++i) printf("active_var[%d]=%s\n", i, active_vars[i]);

    printf("\n");

    for (ind in term_wt)
    {
        if (term_wt[ind] == 0) continue;

        n_inds = split(ind, ind_arr, SUBSEP);

        for (i=1; i<=n_inds; ++i)
            varstr = substr(varstr, 1, max_var - ind_arr[i]) "1" substr(varstr, max_var - ind_arr[i] + 2, ind_arr[i]);

        printf("weight[\"%s\"]=%s\n", varstr, term_wt[ind]);

        for (i=1; i<=n_inds; ++i)
            varstr = substr(varstr, 1, max_var - ind_arr[i]) "2" substr(varstr, max_var - ind_arr[i] + 2, ind_arr[i]);
    }

    printf("\n");

    penalty = "";

    for (ind in term_wt)
    {
        if (term_wt[ind] == 0) continue;

        s = sprintf("%d", term_wt[ind]);

        n_inds = split(ind, ind_arr, SUBSEP);

        for (i=1; i<=n_inds; ++i) s = s sprintf("*%s", active_vars[ind_arr[i]]);

        printf("term: %s\n", s);

        penalty = penalty " + " s;
    }

    printf("\npenalty: %s\n", penalty);
}

END { 

    if (debug)
    {
	printf("constant=%d\n", constant);

	for (v in active_vars) printf("active_var[%d]=%s\n", v, active_vars[v]);

	for (i in linear) printf("linear[%d]=%s\n", i, linear[i]);

	for (i in quadratic) printf("quadratic[%d]=%s\n", i, quadratic[i]);

	for (p in pair) printf("pair[%s]=%s\n", p, pair[p]);
    }

    exit exit_status;
}

' $1 $2
