#!/bin/bash

######################################################################
# Proprietary Information D-Wave Systems Inc.                        #
# Copyright (c) 2017 by D-Wave Systems Inc. All rights reserved.     #
# D-Wave Systems Inc., 3033 Beta Ave., Burnaby, BC, V5G 4M9, Canada. #
######################################################################

################################################################
#
# This shell script applies graph and gauge transformations to either
# QMI or SOL files.  It is not intended to be invoked directly by the
# end user.  Instead, the "dw" command will dispatch to this utility
# when the "trans" subcommand is used.  Transformations are applied in
# the "forward" direction to QMI files and in the "backward" direction
# to SOL files.  Here's what this means.  Suppose we have a QMI file
# named in.qmi and we want to transform it before executing.  We can
# do this:
#
#    in.qmi ==(dw-trans.bash)==> *in*.qmi
#
#                                   |
#                                execute
#                                   |
#                                   V
#
#    in.sol <==(dw-trans.bash)== *in*.sol
#
# The forward transformed QMI is *in*.qmi and we can execute this QMI
# to obtain *in*.sol.  Use dw-trans.bash to backward transform
# *in*.sol to give in.sol.  Then the samples appearing in in.sol are
# equivalent to the samples that would have been obtained by executing
# in.qmi directly.
#
# Another way to express this is via objective values.  Given *any*
# bit-vector for the active qubits in *in*.qmi, we can evaluate the
# objective against the QMI *in*.qmi.  Alternatively, we can transform
# the bit-vector with dw-trans.bash to get the same bit-vector in the
# original (pre-transformed) basis and then evaluate its objective
# against the QMI in.qmi.  These two objectives will match, provided
# that we add the constant offset (generated and displayed by
# dw-trans.bash) to the transformed problem.  More succinctly, after
# including the constant offset, the entire spectrum (including
# degeneracies) of the transformed problem should match the spectrum
# of the original problem.
#
# The structure of this program is a three stage pipeline:
#
#    1) List geometry, transform and QMI or SOL ==>
#
#    2) Generate text representation of transformed QMI or SOL ==>
#
#    3) Re-generate binary representation of QMI or SOL
#
# The first and last stages of the pipeline are Bash shell and the
# second stage is an awk program.
#
# Transform files are simple format text files, containing lines like
# this for non-gauge-transformed qubit mappings:
#
#    Q0000 <== Q0001
#
# and lines like this for gauge-transformed qubit mappings:
#
#    Q0000 <== ~Q0001
#
# It is not necessary to list all qubits in the current geometry -
# only those qubits appearing in the QMI or SOL being transformed.
# The order of lines in a transform file does not matter.  Here is a
# complete transform file which permutes the first four qubits of the
# upper-left unit cell and applies three gauge transformations at the
# same time:
#
# Q0001 <== Q0000
# Q0003 <== Q0001
# Q0002 <== ~Q0002
# Q0000 <== Q0003
# Q0004 <== Q0004
# Q0005 <== ~Q0005
# Q0006 <== ~Q0006
# Q0007 <== Q0007
#
################################################################

shopt -s extglob

usage="dw-trans.bash transform.t {trans.qmi orig.qmi | trans.sol orig.sol}"

trans_file=
orig_qmi=
trans_qmi=
orig_sol=
trans_sol=

# **************** Command line argument processing ****************

for arg
do
    if [[ $arg == *.t ]]
    then
	if [[ -n $trans_file ]]
	then
	    echo "USAGE: $usage"
	    exit 1
	fi
	trans_file=$arg
    fi

    if [[ $arg == *.qmi ]]
    then
	qmi_mode=1
	if [[ -n $orig_qmi ]]
	then
	    echo "USAGE: $usage"
	    exit 1
	else
	    [[ -n $trans_qmi ]] && orig_qmi=${arg} || trans_qmi=${arg}
	fi
    fi

    if [[ $arg == *.sol ]]
    then
	sol_mode=1
	if [[ -n $orig_sol ]]
	then
	    echo "USAGE: $usage"
	    exit 1
	else
	    [[ -n $trans_sol ]] && orig_sol=${arg} || trans_sol=${arg}
	fi
    fi
done

if [[ -n $qmi_mode && -n $sol_mode ]]
then
    echo "USAGE: $usage"
    exit 1
fi

if [[ -z $qmi_mode && -z $sol_mode ]]
then
    echo "USAGE: $usage"
    exit 1
fi

if [[ -n $trans_qmi && -z $orig_qmi ]]
then
    echo "USAGE: $usage"
    exit 1
fi

if [[ -n $trans_sol && -z $orig_sol ]]
then
    echo "USAGE: $usage"
    exit 1
fi

# **************** Finished command line argument processing ****************

if [[ -n $sol_mode ]]
then
    ws_dir=$(dw get path | awk '/VALUE/ {sub(/.*=/,""); gsub(/'"'"'/,""); print $0}')
    orig_sol_file=${ws_dir}/${orig_sol}
    trans_sol_file=${ws_dir}/${trans_sol}
fi

# **************** First pipeline stage begins here ****************

# Dump the geometry file, the .t file and either the .qmi or .sol
# into awk via a shell pipeline:

{
    cat $DW_INTERNAL__WORKSPACE/geometry
    echo "READ TRANSFORM"
    cat $trans_file

    if [[ -n $qmi_mode ]]
    then
	echo "READ QMI"
	dw get $orig_qmi
    fi

    if [[ -n $sol_mode ]]
    then
	echo "READ SOL"
	dw get $orig_sol
    fi
} |

# **************** AWK program begins here ****************

awk '

BEGIN { read_geometry = 1; debug = 0; }

read_geometry == 1 && /^L:/ { DW_L = $2 }
read_geometry == 1 && /^M:/ { DW_M = $2 }
read_geometry == 1 && /^N:/ { DW_N = $2 }

read_geometry == 1 && /^Q/ {
       gsub(" ", "");
       split($0, token, ":");
       qubit_range = token[1];
       bit_mask = token[2];
       sub("-.*$", "", qubit_range);
       sub("^Q", "", qubit_range);
       qubit_index = 0 + qubit_range;
       bit_index = 1;
       bits = length(bit_mask);

       while (bit_index <= bits)
       {
           qubit[qubit_index] = substr(bit_mask, bit_index, 1);
           ++ qubit_index;
           ++ bit_index;
       }
}

read_geometry == 1 && /^C/ {
       gsub(" ", "");
       split($0, token, ":");
       coupler_range = token[1];
       bit_mask = token[2];
       sub("-.*$", "", coupler_range);
       sub("^C", "", coupler_range);
       coupler_index = 0 + coupler_range;
       bit_index = 1;
       bits = length(bit_mask);

       while (bit_index <= bits)
       {
           coupler[coupler_index] = substr(bit_mask, bit_index, 1);
           ++ coupler_index;
           ++ bit_index;
       }
}

read_qmi == 1 && $1 ~ /^Q[0-9]*$/ {
       value = $0;
       sub(/.*<==/, "", value);
       gsub(/ /, "", value);

       q = $0;
       sub(/<==.*/, "", q);
       gsub(/ /, "", q);

       orig_qmi[q] = value;
}

read_qmi == 1 && $1 ~ /^C[0-9]*$/ {
       value = $0;
       sub(/.*<==/, "", value);
       gsub(/ /, "", value);

       c = $0;
       sub(/<==.*/, "", c);
       gsub(/ /, "", c);

       orig_qmi[c] = value;
}

read_qmi == 1 && $1 ~ /^CONST$/ {
       const_offset = $3;
}

read_qmi == 1 && $1 ~ /^SCALE_FACTOR$/ {
       scale_factor = $3;
}

read_sol == 1 && NF == 3 && $1 ~ /^Q/ {
       q = $0;
       sub(/<==.*/, "", q);
       gsub(/ /, "", q);
       q = 0 + substr(q, 2);
       gv = gauge[T_inv[q]];
       printf "Q%04d <== %s\r\n", T_inv[q], gv + $3 - 2 * gv * $3;
}

read_sol == 1 && NF == 4 && $2 ~ /^SOLUTION$/ {
       print $0;
}

read_sol == 1 && /VALUE/ && /END/ {
       printf "END SOLUTION\r\n";
}

read_transform == 1 && /READ QMI/ { read_transform = 0; read_qmi = 1; }

read_transform == 1 && /READ SOL/ { read_transform = 0; read_sol = 1; }

read_transform == 1 {
       x = $0;
       sub(/.*<==/, "", x);
       gsub(/ /, "", x);
       gsub(/Q/, "", x);
       tilde = gsub(/~/, "", x);
       x = 0 + x;
       if (tilde == 0) gauge[x] = 0;
       else            gauge[x] = 1;

       y = $0;
       sub(/<==.*/, "", y);
       gsub(/ /, "", y);
       gsub(/Q/, "", y);
       y = 0 + y;

       if (x in T) {
           printf "Qubit Q%04d appears multiple times in domain(T)\r\n", x;
           transform_error = 1;
       }

       if (y in T_inv) {
           printf "Qubit Q%04d appears multiple times in range(T)\r\n", y;
           transform_error = 1;
       }

       T[x] = y;
       T_inv[y] = x;
}

read_geometry == 1 && /READ TRANSFORM/ { read_geometry = 0; read_transform = 1; }

END {
    qubits = 2*DW_L*DW_M*DW_N;

    couplers = DW_L*DW_L*DW_M*DW_N + DW_L*DW_M*(DW_N-1) + DW_L*(DW_M-1)*DW_N;

    if (debug) {
        printf "CONST:%d\n", const_offset > "/dev/stderr"

	for (i=0; i<qubits; ++i) {
	    printf "Q%04d:%d ", i, qubit[i] > "/dev/stderr";
	    if (i%16 == 15) printf "\n" > "/dev/stderr";
	}

	for (i=0; i<couplers; ++i) {
	    printf "C%04d:%d ", i, coupler[i] > "/dev/stderr";
	    if (i%16 == 15) printf "\n" > "/dev/stderr";
	}

	for (x in T) {
	    printf "Q%04d = T[Q%04d]\r\n", T[x], x > "/dev/stderr";
	}

	for (y in T_inv) {
	    printf "Q%04d = T_inv[Q%04d]\r\n", T_inv[y], y > "/dev/stderr";
	}

	for (i=0; i<qubits; ++i) {
	    printf "gauge[Q%04d]=%d\r\n", i, gauge[i] > "/dev/stderr";
	}
    }

# First apply the gauge transformation:

    constant = const_offset;

    for (o in orig_qmi) {
        if (o ~ /^Q/) {
            q = 0 + substr(o, 2);
            if (gauge[q] == 0) {
                gauge_qmi[o] += orig_qmi[o];
            } else if (gauge[q] == 1) {
                constant += orig_qmi[o];
                gauge_qmi[o] -= orig_qmi[o];
            }
        }
        if (o ~ /^C/) {
            c = 0 + substr(o, 2);
            Qubits_from_Coupler(c);
            q1label = sprintf("Q%04d", pq1);
            q2label = sprintf("Q%04d", pq2);
            if (gauge[pq1] == 0 && gauge[pq2] == 0) {
                gauge_qmi[o] = orig_qmi[o];
            } else if (gauge[pq1] == 1 && gauge[pq2] == 0) {
                gauge_qmi[q2label] += orig_qmi[o];
                gauge_qmi[o] = -orig_qmi[o];
            } else if (gauge[pq1] == 0 && gauge[pq2] == 1) {
                gauge_qmi[q1label] += orig_qmi[o];
                gauge_qmi[o] = -orig_qmi[o];
            } else if (gauge[pq1] == 1 && gauge[pq2] == 1) {
                constant += orig_qmi[o];
                gauge_qmi[q1label] -= orig_qmi[o];
                gauge_qmi[q2label] -= orig_qmi[o];
                gauge_qmi[o] = orig_qmi[o];
            }
        }
    }

# If we handled a QMI, output the scale factor and constant:

    if (read_qmi)
    {
        printf "SCALE_FACTOR <== %g\r\n", scale_factor;
        printf "CONST <== %g\r\n", constant;
    }

# Now apply the gauge transformation and output the weights and strengths:

    for (o in gauge_qmi) {
        if (o ~ /^Q/) {
            q = 0 + substr(o, 2);
            printf "Q%04d <== %s\r\n", T[q], gauge_qmi[o];
        }
        if (o ~ /^C/) {
            c = 0 + substr(o, 2);
            Qubits_from_Coupler(c);
            tpq1 = T[pq1];
            tpq2 = T[pq2];
            tc = Coupler_from_Qubits(tpq1,tpq2);
            printf "C%04d <== %s\r\n", tc, gauge_qmi[o];
        }
    }
}

function ChimInd_from_Linear(linear)
{
    k = (linear % DW_L);
    linear -= k;
    linear /= DW_L;

    u = (linear % 2);
    linear -= u;
    linear /= 2;

    j = (linear % DW_N);
    linear -= j;
    linear /= DW_N;

    i = linear;
}

function Linear_from_ChimInd(i,j,u,k,temp)
{
    return 2*DW_N*DW_L*i + 2*DW_L*j + DW_L*u + k;
}

function Coupler_from_Qubits(pqubit1, pqubit2)
{
    ChimInd_from_Linear(pqubit1);
    i1 = i;
    j1 = j;
    u1 = u;
    k1 = k;

    ChimInd_from_Linear(pqubit2);
    i2 = i;
    j2 = j;
    u2 = u;
    k2 = k;

    # Intracell coupler:
    if (i1 == i2 && j1 == j2 && u1 != u2)
    {
        if (u1 == 0) offset = k1 * DW_L + k2;
        if (u2 == 0) offset = k2 * DW_L + k1;
        return i1 * (DW_L*DW_L*DW_N) + j1 * (DW_L*DW_L) + offset;
    }

    # Horizontal Intercell coupler:
    if (i1 == i2 && ((j1 - j2 == 1) || (j1 - j2 == -1)) && u1 == 1 && u2 == 1 && k1 == k2)
    {
        if (j1 < j2) jmin = j1;
        if (j2 < j1) jmin = j2;
        return DW_L*DW_L*DW_M*DW_N + i1 * (DW_L*(DW_N - 1)) + jmin * DW_L + k1;
    }

    # Vertical Intercell coupler:
    if (((i1 - i2 == 1) || (i1 - i2 == -1)) && j1 == j2 && u1 == 0 && u2 == 0 && k1 == k2)
    {
        if (i1 < i2) imin = i1;
        if (i2 < i1) imin = i2;
        return DW_L*DW_L*DW_M*DW_N + DW_L*DW_M*(DW_N - 1) + imin * (DW_L*DW_N) + j1 * DW_L + k1;
    }

    printf "error: qubits %d and %d are not connected by a coupler\n", pqubit1, pqubit2;
    exit 1;
}

function Qubits_from_Coupler(coupler, temp_coupler, offset, ti, tj, tk0, tk1)
{
    temp_coupler = coupler;

    # Intracell coupler:
    if (coupler < DW_L*DW_L*DW_M*DW_N)
    {
        offset = coupler % (DW_L * DW_L);
        coupler -= offset;
        coupler /= (DW_L * DW_L);

        tj = coupler % DW_N;
        coupler -= tj;
        coupler /= DW_N;

        ti = coupler;

        tk1 = offset % DW_L;
        offset -= tk1;
        tk0 = offset / DW_L;

        pq1 = Linear_from_ChimInd(ti, tj, 0, tk0);
        pq2 = Linear_from_ChimInd(ti, tj, 1, tk1);

        return;
    }

    coupler = temp_coupler - DW_L*DW_L*DW_M*DW_N;

    # Horizontal Intercell coupler:
    if (coupler < DW_L*DW_M*(DW_N - 1))
    {
        tk1 = coupler % DW_L;
        coupler -= tk1;
        coupler /= DW_L;

        tj = coupler % (DW_N - 1);
        coupler -= tj;
        coupler /= (DW_N - 1);

        ti = coupler;

        pq1 = Linear_from_ChimInd(ti,   tj, 1, tk1);
        pq2 = Linear_from_ChimInd(ti, 1+tj, 1, tk1);

        return;
    }

    coupler = temp_coupler - DW_L*DW_L*DW_M*DW_N - DW_L*DW_M*(DW_N - 1);

    # Vertical Intercell coupler:
    if (coupler < DW_L*(DW_M - 1)*DW_N)
    {
        tk0 = coupler % DW_L;
        coupler -= tk0;
        coupler /= DW_L;

        tj = coupler % DW_N;
        coupler -= tj;
        coupler /= DW_N;

        ti = coupler;

        pq1 = Linear_from_ChimInd(  ti, tj, 0, tk0);
        pq2 = Linear_from_ChimInd(1+ti, tj, 0, tk0);

        return;
    }

    printf "error: coupler %d out of range\n", temp_coupler;
    exit 1;
}

function CouplerType(coupler, temp_coupler)
{
    temp_coupler = coupler;

    if (coupler < DW_L*DW_L*DW_M*DW_N) return "intracell";

    coupler -= DW_L*DW_L*DW_M*DW_N;

    if (coupler < DW_L*DW_M*(DW_N - 1)) return "horizontal";

    coupler -= DW_L*DW_M*(DW_N - 1);

    if (coupler < DW_L*(DW_M - 1)*DW_N) return "vertical";

    printf "error: coupler %d out of range\n", temp_coupler;
    exit 1;
}

' | tr -d '\r' |

# **************** Third pipeline stage begins here ****************

# Convert the text representation of the QMI or SOL into its binary
# representation and write it to disk.

{
    # QMI files can be converted from text representation to binary
    # easily with "dw set *.qmi", which reads from its standard input.

    if [[ -n $qmi_mode ]]
    then
	dw set ${trans_qmi}
    elif [[ -n $sol_mode ]]
    
    # SOL files have to be constructed by hand for now.  We can actually
    # write out binary files using printf in the shell.

    then
	declare -a qv
	qubits=$(dw get geometry | awk '$1 ~ /qubits/ {sub(/.*=/,"",$2); print $2}')
	{
	    head -c20 ${orig_sol_file}
	    while read token1 token2 token3
	    do
		[[ $token1 != END && $token2 == SOLUTION ]] && num_solutions=${token3%% *}

		if [[ $token2 == SOLUTION && ( $num_solutions != 1 || $token1 == END ) ]]
		then
		    for ((i=0; i<qubits; ++i))
		    do
			[[ -n ${qv[i]} ]] && val=${qv[i]} || val=3
			printf "\\""${val}\0\0\0"
		    done
		elif [[ $token2 == '<==' ]]
		then
		    qubit_index=${token1##Q*(0)}
		    qv[qubit_index]=${token3}
		fi
	    done
	    tail -c$(( (8 + 4) * num_solutions )) ${orig_sol_file}
	} > ${trans_sol_file}
    fi
}
